from typing import List

from bibx._entities.article import Article
from bibx._entities.collection import Collection
from bibx._entities.collection_builders.base import CollectionBuilder


class SimpleCollectionBuilder(CollectionBuilder):
    def __init__(self, articles: List[Article]):
        self.articles = articles

    def build(self) -> Collection:
        return Collection(self.articles[:])



from bibx._entities.collection import Collection
from bibx._entities.collection_builders.base import CollectionBuilder


class CrossRefCollectionBuilder(CollectionBuilder):
    def __init__(self, query: str, count: int = 100):
        self._query = query
        self._count = count

    def with_count(self, count: int):
        self._count = count
        return self

    def build(self) -> Collection:
        return Collection([])



import networkx as nx

from bibx.algorithms.sap import Sap


def create_toy_graph() -> nx.DiGraph:
    r"""
    Creates a toy graph with known sap.

    The toy graph has the following shape:

    a   b   c
      \ | /
        d
      / | \
    e   f   g
    """

    g = nx.DiGraph()
    for n in "abcdefg":
        # we need to have years for all the nodes
        g.add_node(n, year=2000)
    for n in "abc":
        g.add_edge(n, "d")
    for n in "efg":
        g.add_edge("d", n)
    return g


def test_sap():
    g = create_toy_graph()
    s = Sap()
    g = s.tree(g)

    # a, b and c should be leaves
    for n in "abc":
        assert g.nodes[n]["leaf"] > 0
        assert g.nodes[n]["trunk"] == 0
        assert g.nodes[n]["root"] == 0

    # e, f, and g should be roots
    for n in "efg":
        assert g.nodes[n]["leaf"] == 0
        assert g.nodes[n]["trunk"] == 0
        assert g.nodes[n]["root"] > 0

    # node d should be the trunk
    assert g.nodes["d"]["leaf"] == 0
    assert g.nodes["d"]["trunk"] > 0
    assert g.nodes["d"]["root"] == 0



import io

from bibx._entities.collection_builders.scopus import ScopusCollectionBuilder


def test_scopus_works():
    file = io.StringIO(
        """
@ARTICLE{Steblinski2022,
author={Steblinski, P. and Blachowicz, T. and Ehrmann, A.},
title={Analysis of the energy distribution of iron nano-spheres for bit-patterned media},
journal={Journal of Magnetism and Magnetic Materials},
year={2022},
volume={562},
doi={10.1016/j.jmmm.2022.169805},
art_number={169805},
note={cited By 0},
url={https://www.scopus.com/inward/record.uri?eid=2-s2.0-85135886322&doi=10.1016%2fj.jmmm.2022.169805&partnerID=40&md5=759122e06fc4fd8cc4a87011561de73b},
affiliation={Koszalin University of Technology, Faculty of Electronics and Informatics, Koszalin, Poland; Silesian University of Technology, Institute of Physics – Center for Science and Education, Gliwice, Poland; Bielefeld University of Applied Sciences, Faculty of Engineering Sciences and Mathematics, Bielefeld, Germany},
abstract={Improved magnetic memory systems belong to the main research topics in spintronics. Here we show micromagnetic simulations used to analyze the energy density of nano-scaled iron spheres. Layers of different thickness, partly coated with iron oxide, were tested in terms of spatial uniformity of the physical system energy. For a single non-coated or iron-oxide coated droplet, the spatial distribution of the total energy is not uniform and depends on the nano-droplet size. Additionally, for systems consisting of four objects, the relation between relative distance and the resultant magnetization distribution was analyzed. The mutual relation between droplet size and the underlying magnetization distribution as well as the character of local energy extrema was investigated. The size changes for the four-droplet system were compared with the single object behavior to obtain a criterion for the minimum distance between spheres to behave as a single object. The calculations revealed that the oxidized spheres could be placed closer to each other in comparison to the non-coated system. For the proposed oxide coated system, the increase of this maximum packing density is equal to about 12%, as compared to the non-coated system. © 2022 Elsevier B.V.},
author_keywords={Ferromagnetism;  Magnetic anisotropy;  Micromagnetism},
keywords={Drops;  Ferromagnetic materials;  Ferromagnetism;  Magnetic anisotropy;  Magnetic storage;  Magnetization;  Spheres, Bit-patterned media;  Coated systems;  Droplets sizes;  Energy distributions;  Magnetic memory;  Magnetization distribution;  Memory systems;  Micromagnetisms;  Nano-spheres;  Single object, Iron oxides},
references={Richter, H., Dobin, A., Heinonen, O., Gao, K., Veerdonk, R., Lynch, R., Xue, J., Brockie, R., Recording on Bit-Patterned Media at Densities of 1 Tb/in and Beyond (2006) IEEE Trans. Magn., 42, pp. 2255-2260; Koltsov, D.K., Adeyeye, A.O., Welland, M.E., Tricker, D.M., Single-domain circular nanomagnets (1999) Phys. Rev. Lett., 83, p. 1042; Zhang, W., Haas, S., Phase diagram of magnetization reversal processes in nanorings (2010) Phys. Rev. B, 81; He, K., Smith, D.J., McCartney, M.R., Effects of vortex chirality and shape anisotropy on magnetization reversal of Co nanorings (invited) (2010) J. Appl. Phys., 107, p. 09D307; Blachowicz, T., Ehrmann, A., Square nano-magnets as bit-patterned media with doubled possible data density (2017) Mater. Today:. Proc., 4, pp. S226-S231; Döpke, C., Grothe, T., Steblinski, P., Klöcker, M., Sabantina, L., Kosmalska, D., Blachowicz, T., Ehrmann, A., Magnetic Nanofiber Mats for Data Storage and Transfer (2019) Nanomaterials, 9, p. 92; Steblinski, P., Blachowicz, T., Conception of magnetic memory switched by time dependent current density and current electron spin polarization (2019) International Journal of Electronics and Telecommunications, 65, p. 309; Shutyi, A.M., Sementsov, D.I., Multistability of the Magnetization Reversal of a Nanoparticle with Cubic Anisotropy (2020) JETP Lett., 111, pp. 619-626; Russier, V., De-Montferrand, C., Lalatonne, Y., Motte, L., Magnetization of densely packed interacting magnetic nanoparticles with cubic and uniaxial anisotropies: a Monte Carlo study (2013) J. Appl. Phys., 114; Simeonidis, K., Martinez-Boubeta, C., Serantes, D., Ruta, S., Chubykalo-Fesenko, O., Chantrell, R., Oró-Solé, J., Angelakeris, M., Controlling Magnetization Reversal and Hyperthermia Efficiency in Core−Shell Iron−Iron Oxide Magnetic Nanoparticles by Tuning the Interphase Coupling (2020) ACS Appl. Nano Mater., 3, p. 4465; Nemati, Z., Alonso, J., Khurshid, H., Phan, M.H., Srikanth, H., Core/shell iron/iron oxide nanoparticles: are they promising for magnetic hyperthermia? (2016) RSC Adv., 6, p. 38697; Kim, Y.-W., Park, H.S., Hyun Soon Park Microstructural and Magnetic Characterization of Iron Oxide Nanoparticles Fabricated by Pulsed Wire Evaporation (2019) Electron. Mater. Lett., 15 (6), pp. 665-672; Scholz, W., Fidler, J., Schrefl, T., Suess, D., Dittrich, R., Forster, H., Tsiantos, V., Scalable parallel micromagnetic solvers for magnetic nanostructures (2003) Comput. Mater. Sci., 28, p. 366; (2019), TAO Users Manual, ANL/MCS-TM-322 Rev. 3.11, Argonne National Laboratory; Kostopoulou, A., Brintakis, K., Vasilakaki, M., Trohidou, K.N., Douvalis, A.P., Lascialfari, A., Manna, L., Lappas, A., Assembly-mediated Interplay of Dipolar Interactions and Surface Spin Disorder in Colloidal Maghemite Nanoclusters (2014) Nanoscale, 6 (7), pp. 3764-3776; Coey, J.M.D., Magnetism and Magnetic Materials (2010), Cambridge University Press; Cullity, B.D., Graham, C.D., Introduction to magnetic materials (2008), 2nd ed. Wiley; Johansson, C., Hanson, M., Pedersen, M.S., Morup, S., Magnetic properties of magnetic liquids with iron-oxide particles—the influence of anisotropy and interactions (1997) J. Magn. Magn. Mater., 173, pp. 5-14; Aguilera-del-Toro, R.H., Aguilera-Granja, F., Torres, M.B., Vega, A., Relation between structural patterns and magnetism in small iron oxide clusters: reentrance of the magnetic moment at high oxidation ratios (2021) Phys. Chem. Chem. Phys., 23, pp. 246-272; Erlebach, A., Hühn, C., Jana, R., Sierka, M., Structure and magnetic properties of (Fe2O3)n clusters (n = 1–5) (2014) Phys. Chem. Chem. Phys., 16 (48), pp. 26421-26426},
correspondence_address1={Blachowicz, T.; Silesian University of Technology, S. Konarskiego 22B str., 44-100 Gliwice, Poland; email: tomasz.blachowicz@polsl.pl},
publisher={Elsevier B.V.},
issn={03048853},
coden={JMMMD},
language={English},
abbrev_source_title={J Magn Magn Mater},
document_type={Article},
source={Scopus},
}
"""
    )
    data = ScopusCollectionBuilder(file).build()
    assert len(data.articles) == 1
    (article,) = data.articles
    assert (
        article.title
        == "Analysis of the energy distribution of iron nano-spheres for bit-patterned media"
    )
    assert len(list(data.citation_pairs)) == 20



import json
import re
from contextlib import suppress
from typing import Iterable, Optional, TextIO

import bibtexparser

from bibx._entities.article import Article
from bibx._entities.collection import Collection
from bibx._entities.collection_builders.base import CollectionBuilder
from bibx.exceptions import MissingCriticalInformation


class ScopusCollectionBuilder(CollectionBuilder):
    def __init__(self, *scopus_files: TextIO):
        self._files = scopus_files
        for file in self._files:
            file.seek(0)

    def build(self) -> Collection:
        articles = self._get_articles_from_files()
        return Collection(list(articles))

    def _get_articles_from_files(self) -> Iterable[Article]:
        for file in self._files:
            db = bibtexparser.load(file)
            for entry in db.entries:
                with suppress(MissingCriticalInformation):
                    yield self._article_from_entry(entry)

    def _article_from_entry(self, entry: dict) -> Article:
        if "author" not in entry or "year" not in entry:
            raise MissingCriticalInformation()
        return Article(
            authors=entry["author"].split(" and "),
            year=int(entry["year"]),
            title=entry.get("title"),
            journal=entry.get("journal"),
            volume=entry.get("volume"),
            issue=entry.get("issue"),
            page=entry.get("art_number"),
            doi=entry.get("doi"),
            references=list(self._articles_from_references(entry.get("references"))),
            keywords=entry.get("keywords", "").split("; "),
            extra=entry,
            sources={json.dumps(entry)},
        )

    def _articles_from_references(self, references: Optional[str]) -> Iterable[Article]:
        if references is None:
            references = ""
        for reference in references.split("; "):
            with suppress(MissingCriticalInformation):
                yield self._article_from_reference(reference)

    @staticmethod
    def _article_from_reference(reference: str) -> Article:
        match = re.search(r"\((\d{4})\)", reference)
        if not match:
            raise MissingCriticalInformation()
        year = int(match.groups()[0])
        author = reference.split(",", maxsplit=2)[0].strip()
        doi = re.search(r"(10.\d{4,9}/[-._;()/:A-Z0-9]+)", reference)
        return Article(
            authors=[author],
            year=year,
            _label=reference,
            doi=doi.groups()[0] if doi else None,
            sources={reference},
        )



class BibXError(Exception):
    """
    Any exception known by wostools.
    """


class InvalidIsiLineError(BibXError, ValueError):
    """
    Raised when we encounter an invalid line when processing an ISI file.
    """

    def __init__(self, line: str):
        super().__init__(f"'{line}' is not a valid ISI file line")


class InvalidIsiReference(BibXError, ValueError):
    """
    Raised when we try to create an article out of an invalid ISI reference.
    """

    def __init__(self, reference: str):
        super().__init__(f"{reference} does not look like an ISI citation")


class MissingCriticalInformation(BibXError, ValueError):
    """
    Raised when we don't have the publication year of an article.
    """

    def __init__(self):
        super().__init__("Article is missing some critical information")



from typing import TextIO

from bibx._entities.article import Article
from bibx._entities.collection import Collection
from bibx._entities.collection_builders.isi import IsiCollectionBuilder
from bibx._entities.collection_builders.scopus import ScopusCollectionBuilder

__all__ = ["Article", "Collection", "read_scopus", "read_wos"]

__version__ = "0.0.1a1"


def read_scopus(*files: TextIO) -> Collection:
    """
    Takes any number of bibtex files from scopus and generates a collection.

    :param files: Scopus bib files open.
    :return: the collection
    """
    return ScopusCollectionBuilder(*files).build()


def read_wos(*files: TextIO) -> Collection:
    """
    Takes any number of wos text files and returns a collection.

    :param files: WoS files open.
    :return: the collection
    """
    return IsiCollectionBuilder(*files).build()



from dataclasses import dataclass, field
from typing import List, Mapping, Optional, Set


@dataclass
class Article:
    authors: List[str]
    year: int
    title: Optional[str] = None
    journal: Optional[str] = None
    volume: Optional[str] = None
    issue: Optional[str] = None
    page: Optional[str] = None
    doi: Optional[str] = None
    _label: Optional[str] = None
    references: Optional[List["Article"]] = field(default_factory=list)
    keywords: Optional[List[str]] = field(default_factory=list)
    sources: Optional[Set[str]] = field(default_factory=set)
    extra: Optional[Mapping] = field(default_factory=dict)

    @property
    def key(self):
        author = self.authors[0].split(" ")[0].replace(",", "")
        year = self.year
        return f"{author}{year}".lower()

    @property
    def label(self):
        if self._label is not None:
            return self._label
        pieces = {
            "AU": self.authors[0].replace(",", ""),
            "PY": str(self.year),
            "J9": str(self.journal),
            "VL": f"V{self.volume}" if self.volume else None,
            "BP": f"P{self.page}" if self.page else None,
            "DI": f"DOI {self.doi}" if self.doi else None,
        }
        return ", ".join(value for value in pieces.values() if value)



import logging
from dataclasses import dataclass
from typing import Iterable, List, Tuple

from bibx._entities.article import Article

logger = logging.getLogger(__name__)


@dataclass
class Collection:
    articles: List[Article]

    @property
    def all_articles(self) -> Iterable[Article]:
        """
        Yields all articles and references

        :return: an iterable over `Article`.
        """
        cache = {article.key: article for article in self.articles}
        for article in self.articles:
            for reference in article.references:
                cache.setdefault(reference.key, reference)
        yield from cache.values()

    @property
    def citation_pairs(self) -> Iterable[Tuple[Article, Article]]:
        cache = {article.key: article for article in self.articles}
        for article in self.articles:
            if not article.references:
                continue
            for reference in article.references:
                # Yield a full article if we have the metadata for a citation
                if article.key != reference.key and reference.key in cache:
                    logger.debug("Found a cache hit for key %s", reference.key)
                    yield article, cache[reference.key]
                else:
                    yield article, reference

    def merge(self, other: "Collection") -> "Collection":
        """
        Creates a new collection merging the articles by key.

        :param other: collection to merge to.
        :return: a new collection object.
        """
        keys = {a.key for a in self.articles}
        merged = self.articles[:]
        merged.extend(a for a in other.articles if a.key not in keys)
        return Collection(merged)



import dataclasses
import logging
from typing import List, Set, Tuple

import networkx as nx
from networkx.algorithms.community.louvain import louvain_communities

from bibx import Article, Collection

YEAR = "year"
LEAF = "leaf"
ROOT = "root"
TRUNK = "trunk"
BRANCH = "branch"
SAP = "_sap"
LEAF_CONNECTIONS = "_leaf_connections"
ELABORATE_SAP = "_elaborate_sap"
ROOT_CONNECTIONS = "_root_connections"
RAW_SAP = "_raw_sap"

logger = logging.getLogger(__name__)


def _limit(attribute: List[Tuple[str, int]], _max: int):
    if _max is not None:
        sorted_attribute = sorted(attribute, key=lambda x: x[1], reverse=True)
        attribute = sorted_attribute[:_max]
    return attribute


class Sap:
    def __init__(
        self,
        max_roots: int = 20,
        max_leaves: int = 50,
        max_trunk: int = 20,
        min_leaf_connections: int = 3,
        max_leaf_age: int = 7,
        max_branch_size: int = 15,
    ):
        """
        Create a Sap instance with the given parameters.

        :param max_roots: maximum number of roots on the tree
        :param max_leaves: maximum number of leaves on the tree
        :param max_trunk: maximum number of trunk nodes in the tree
        :param min_leaf_connections: minimum number of connections between leaves and roots
        :param max_leaf_age: maximum age for a leaf
        """
        self.max_roots = max_roots
        self.max_leaves = max_leaves
        self.max_trunk = max_trunk
        self.min_leaf_connections = min_leaf_connections
        self.max_leaf_age = max_leaf_age
        self.max_branch_size = max_branch_size

    def _compute_root(self, graph: nx.DiGraph) -> nx.DiGraph:
        """
        Takes in a connected graph and returns it labeled with a `root` property.
        :return: Labeled graph with the root property.
        """
        g = graph.copy()
        valid_roots = [
            (n, g.in_degree(n)) for n in g.nodes if g.out_degree(n) == 0  # noqa
        ]
        sorted_roots = _limit(valid_roots, self.max_roots)
        nx.set_node_attributes(g, 0, ROOT)
        for (node, degree) in sorted_roots:
            g.nodes[node][ROOT] = degree
        return g

    def _compute_leaves(self, graph: nx.DiGraph) -> nx.DiGraph:
        """
        Takes in a connected graph and returns it labeled with a `leaf` property.
        :param graph: Connected and filtered graph to work with.
        :return: Labeled graph with the leaf property.
        """
        g = graph.copy()
        try:
            roots = [n for n, d in g.nodes(data=True) if d[ROOT] > 0]
        except AttributeError:
            raise TypeError("It's necessary to have some roots")
        if not roots:
            raise TypeError("It's necessary to have some roots")

        nx.set_node_attributes(g, 0, ROOT_CONNECTIONS)
        for node in roots:
            g.nodes[node][ROOT_CONNECTIONS] = 1
        topological_order = list(nx.topological_sort(g))
        for node in reversed(topological_order):
            nbs = list(g.successors(node))
            if nbs:
                g.nodes[node][ROOT_CONNECTIONS] = sum(
                    g.nodes[n][ROOT_CONNECTIONS] for n in nbs
                )

        potential_leaves = [
            (node, g.nodes[node][ROOT_CONNECTIONS])
            for node in g.nodes
            if g.in_degree(node) == 0  # noqa
        ]
        extended_leaves = potential_leaves[:]

        if self.min_leaf_connections is not None:
            potential_leaves = [
                (n, c) for n, c in potential_leaves if c >= self.min_leaf_connections
            ]

        if self.max_leaf_age is not None:
            potential_leaves = [
                (n, c) for n, c in potential_leaves if YEAR in g.nodes[n]
            ]
            newest_year = max(
                g.nodes[n][YEAR] for n, _ in potential_leaves if g.nodes[n][YEAR]
            )
            earliest_publication_year = newest_year - self.max_leaf_age
            potential_leaves = [
                (n, c)
                for n, c in potential_leaves
                if g.nodes[n][YEAR] >= earliest_publication_year
            ]

        if not potential_leaves:
            logger.info(
                "Reverting leaf cut policies, as they remove all possible leaves"
            )
            potential_leaves = extended_leaves

        potential_leaves = _limit(potential_leaves, self.max_leaves)
        nx.set_node_attributes(g, 0, LEAF)
        for node, c in potential_leaves:
            g.nodes[node][LEAF] = c
        return g

    @staticmethod
    def _compute_sap(graph: nx.DiGraph) -> nx.DiGraph:
        """
        Computes the sap of each node.
        """
        g = graph.copy()
        try:
            valid_root = [n for n, d in g.nodes(data=True) if d[ROOT] > 0]
            valid_leaves = [n for n, d in g.nodes(data=True) if d[LEAF] > 0]
        except AttributeError:
            raise TypeError("The graph needs to have a 'root' and a 'leaf' attribute")
        if not valid_root or not valid_leaves:
            raise TypeError("The graph needs to have at least some roots and leafs")

        nx.set_node_attributes(g, 0, RAW_SAP)
        nx.set_node_attributes(g, 0, ROOT_CONNECTIONS)
        for node in valid_root:
            g.nodes[node][RAW_SAP] = g.nodes[node][ROOT]
            g.nodes[node][ROOT_CONNECTIONS] = 1
        for node in reversed(list(nx.topological_sort(g))):
            neighbors = list(g.successors(node))
            if neighbors:
                for attr in (RAW_SAP, ROOT_CONNECTIONS):
                    g.nodes[node][attr] = sum(g.nodes[nb][attr] for nb in neighbors)

        nx.set_node_attributes(g, 0, ELABORATE_SAP)
        nx.set_node_attributes(g, 0, LEAF_CONNECTIONS)
        for node in valid_leaves:
            g.nodes[node][ELABORATE_SAP] = g.nodes[node][LEAF]
            g.nodes[node][LEAF_CONNECTIONS] = 1
        for node in nx.topological_sort(g):
            neighbors = list(g.predecessors(node))
            if neighbors:
                for attr in (ELABORATE_SAP, LEAF_CONNECTIONS):
                    g.nodes[node][attr] = sum(g.nodes[nb][attr] for nb in neighbors)

        nx.set_node_attributes(g, 0, SAP)
        for node in g.nodes:
            g.nodes[node][SAP] = (
                g.nodes[node][LEAF_CONNECTIONS] * g.nodes[node][RAW_SAP]
                + g.nodes[node][ROOT_CONNECTIONS] * g.nodes[node][ELABORATE_SAP]
            )

        return g

    def _compute_trunk(self, graph: nx.DiGraph) -> nx.DiGraph:
        """
        Tags leaves.
        """
        g = graph.copy()
        try:
            potential_trunk = [
                (n, g.nodes[n][SAP])
                for n in g.nodes
                if g.nodes[n][ROOT] == 0
                and g.nodes[n][LEAF] == 0
                and g.nodes[n][SAP] > 0
            ]
        except AttributeError:
            raise TypeError(
                "The graph needs to have a 'root', 'leaf' and 'sap' attributes"
            )
        if not potential_trunk:
            raise TypeError("The graph needs to have at least some nodes with sap")

        potential_trunk = _limit(potential_trunk, self.max_trunk)
        nx.set_node_attributes(g, 0, TRUNK)
        for node, sap in potential_trunk:
            g.nodes[node][TRUNK] = sap
        return g

    def _compute_branches(self, graph: nx.DiGraph) -> nx.DiGraph:
        """
        Tags leaves.
        """
        g = graph.copy()
        undirected = g.to_undirected()
        communities: List[Set] = louvain_communities(undirected)
        branches = list(sorted(communities, key=len))[:3]
        nx.set_node_attributes(g, 0, BRANCH)
        for i, branch in enumerate(branches, start=1):
            potential_branch = [(n, g.nodes[n][YEAR]) for n in branch]
            potential_branch = _limit(potential_branch, self.max_branch_size)
            for node, _ in potential_branch:
                g.nodes[node][BRANCH] = i
        return g

    @staticmethod
    def _clear(graph: nx.DiGraph) -> nx.DiGraph:
        """
        Returns a copy of the graph clear of untagged nodes.
        """
        nodes = [
            n
            for n in graph.nodes
            if graph.nodes[n][ROOT] > 0
            and graph.nodes[n][TRUNK] > 0
            and graph.nodes[n][LEAF] > 0
        ]
        return graph.subgraph(nodes)

    @staticmethod
    def create_graph(collection: Collection) -> nx.DiGraph:
        """
        Creates a `networkx.DiGraph` from a `Collection`.

        It uses the article label as a key and adds all the properties of the article to
        the graph.

        :param collection: a `bibx.Collection` instance.
        :return: a `networkx.DiGraph` instance.
        """
        g = nx.DiGraph()
        g.add_edges_from((u.key, v.key) for u, v in collection.citation_pairs)
        for article in collection.articles:
            for reference in article.references:
                _add_article_info(g, reference)
        for article in collection.articles:
            _add_article_info(g, article)
        g.remove_edges_from(nx.selfloop_edges(g))
        return g

    @staticmethod
    def clean_graph(g: nx.DiGraph) -> nx.DiGraph:
        """
        Clean a graph to make it ready for the sap algorithm.

        :param g: graph with unnecessary nodes
        :return: cleaned up giant component
        """
        # Extract the giant component of the graph
        giant_component_nodes = max(nx.weakly_connected_components(g), key=len)
        giant: nx.DiGraph = g.subgraph(giant_component_nodes).copy()

        # Remove nodes that cite one element and are never cited themselves
        giant.remove_nodes_from(
            [
                n
                for n in giant
                if giant.in_degree(n) == 1 and giant.out_degree(n) == 0  # noqa  # noqa
            ]
        )

        # Break loops
        loops = [
            loop for loop in nx.strongly_connected_components(giant) if len(loop) > 1
        ]
        for loop in loops:
            giant.remove_edges_from([(u, v) for u in loop for v in loop])

        return giant

    def tree(self, graph: nx.DiGraph, clear: bool = False) -> nx.DiGraph:
        """
        Computes the whole tree.
        """
        graph = graph.copy()
        graph = self._compute_root(graph)
        graph = self._compute_leaves(graph)
        graph = self._compute_sap(graph)
        graph = self._compute_trunk(graph)
        graph = self._compute_branches(graph)
        if clear:
            graph = self._clear(graph)
        return graph


def _add_article_info(g: nx.DiGraph, article: Article):
    for key, val in dataclasses.asdict(article).items():
        if key in ("sources", "references") or key.startswith("_"):
            continue
        try:
            g.nodes[article.key][key] = val
        except KeyError:
            g.add_node(article.key, **{key: val})



import networkx as nx

from bibx import read_scopus, read_wos
from bibx.algorithms.sap import Sap


def create_toy_graph() -> nx.DiGraph:
    #      a
    #      |
    #      b
    #    / | \
    #   c  d  e
    g = nx.DiGraph()
    for node in "abcde":
        g.add_node(node, year=2000)
    g.add_edge("a", "b")
    for node in "cde":
        g.add_edge("b", node)
    return g


def create_real_graph() -> nx.DiGraph:
    with open("./docs/examples/bit-pattern-savedrecs.txt") as f:
        isi_collection = read_wos(f)
    with open("./docs/examples/scopus.bib") as f:
        scopus_collection = read_scopus(f)
    merged_collection = scopus_collection.merge(isi_collection)
    g = Sap.create_graph(merged_collection)
    g = Sap.clean_graph(g)
    return g


if __name__ == "__main__":
    # graph = create_toy_graph()
    graph = create_real_graph()
    s = Sap()
    graph = s.tree(graph)
    print(graph)



import collections
import functools
import logging
import re
from contextlib import suppress
from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, TextIO

from bibx._entities.article import Article
from bibx._entities.collection import Collection
from bibx._entities.collection_builders.base import CollectionBuilder
from bibx.exceptions import (
    InvalidIsiLineError,
    InvalidIsiReference,
    MissingCriticalInformation,
)

logger = logging.getLogger(__name__)


def _joined(values: List[str], separator: str = " ") -> str:
    return separator.join(value.strip() for value in values)


def _ident(values: List[str]) -> List[str]:
    return [value.strip() for value in values]


def _delimited(values: List[str], delimiter: str = "; ") -> List[str]:
    return [
        word.replace(delimiter.strip(), "")
        for words in values
        for word in words.split(delimiter)
        if word
    ]


def _integer(values: List[str]) -> int:
    if len(values) > 1:
        raise ValueError(f"Expected no more than one item and got {len(values)}")

    first, *_ = values
    return int(first.strip())


@dataclass(frozen=True)
class IsiField:
    key: str
    description: str
    parser: Callable
    aliases: List[str]

    def parse(self, value: List[str]):
        return self.parser(value)


class IsiCollectionBuilder(CollectionBuilder):
    ISI_LINE_PATTERN = re.compile(
        r"^(null|.)?((?P<field>[A-Z0-9]{2})|  )( (?P<value>.*))?$"
    )
    ISI_CITATION_PATTERN = re.compile(
        r"""^(?P<AU>[^,]+),[ ]          # First author
            (?P<PY>\d{4}),[ ]           # Publication year
            (?P<J9>[^,]+)               # Journal
            (,[ ]V(?P<VL>[\w\d-]+))?    # Volume
            (,[ ][Pp](?P<BP>\w+))?      # Start page
            (,[ ]DOI[ ](?P<DI>.+))?     # The all important DOI
            """,
        re.X,
    )

    FIELDS = {
        "AB": IsiField("AB", "Abstract", _joined, ["abstract"]),
        "AF": IsiField("AF", "Author Full Names", _ident, ["author_full_names"]),
        "AR": IsiField("AR", "Article Number", _joined, ["article_number"]),
        "AU": IsiField("AU", "Authors", _ident, ["authors"]),
        "BA": IsiField("BA", "Book Authors", _ident, ["book_authors"]),
        "BE": IsiField("BE", "Editors", _ident, ["editors"]),
        "BF": IsiField(
            "BF", "Book Authors Full Name", _ident, ["book_authors_full_name"]
        ),
        "BN": IsiField(
            "BN",
            "International Standard Book Number (ISBN)",
            _joined,
            ["international_standard_book_number"],
        ),
        "BP": IsiField("BP", "Beginning Page", _joined, ["beginning_page"]),
        "BS": IsiField("BS", "Book Series Subtitle", _joined, ["book_series_subtitle"]),
        "C1": IsiField("C1", "Author Address", _ident, ["author_address"]),
        "CA": IsiField("CA", "Group Authors", _ident, ["group_authors"]),
        "CL": IsiField("CL", "Conference Location", _joined, ["conference_location"]),
        "CR": IsiField(
            "CR",
            "Cited References",
            _ident,
            ["cited_references", "references", "citations"],
        ),
        "CT": IsiField(
            "CT",
            "Conference Title",
            functools.partial(_joined, separator="\n"),
            ["conference_title"],
        ),
        "CY": IsiField("CY", "Conference Date", _joined, ["conference_date"]),
        "DE": IsiField("DE", "Author Keywords", _delimited, ["author_keywords"]),
        "DI": IsiField(
            "DI",
            "Digital Object Identifier (DOI)",
            _joined,
            ["digital_object_identifier", "DOI"],
        ),
        "DT": IsiField("DT", "Document Type", _joined, ["document_type"]),
        "D2": IsiField(
            "D2",
            "Book Digital Object Identifier (DOI)",
            _joined,
            ["book_digital_object_identifier"],
        ),
        "ED": IsiField("ED", "Editors", _ident, ["editors"]),
        "EM": IsiField("EM", "E-mail Address", _ident, ["email_address"]),
        "EI": IsiField(
            "EI",
            "Electronic International Standard Serial Number (eISSN)",
            _joined,
            ["eissn"],
        ),
        "EP": IsiField("EP", "Ending Page", _joined, ["ending_page"]),
        "FU": IsiField(
            "FU",
            "Funding Agency and Grant Number",
            _delimited,
            ["funding_agency_and_grant_number"],
        ),
        "FX": IsiField("FX", "Funding Text", _joined, ["funding_text"]),
        "GA": IsiField(
            "GA",
            "Document Delivery Number",
            _joined,
            ["document_delivery_number"],
        ),
        "GP": IsiField("GP", "Book Group Authors", _ident, ["book_group_authors"]),
        "HO": IsiField("HO", "Conference Host", _joined, ["conference_host"]),
        "ID": IsiField(
            "ID", "Keywords Plus", _delimited, ["keywords_plus", "keywords"]
        ),
        "IS": IsiField("IS", "Issue", _joined, ["issue"]),
        "J9": IsiField(
            "J9",
            "29-Character Source Abbreviation",
            _joined,
            ["source_abbreviation"],
        ),
        "JI": IsiField(
            "JI", "ISO Source Abbreviation", _joined, ["iso_source_abbreviation"]
        ),
        "LA": IsiField("LA", "Language", _joined, ["language"]),
        "MA": IsiField("MA", "Meeting Abstract", _joined, ["meeting_abstract"]),
        "NR": IsiField(
            "NR", "Cited Reference Count", _integer, ["cited_reference_count"]
        ),
        "OI": IsiField(
            "OI",
            "ORCID Identifier (Open Researcher and Contributor ID)",
            _delimited,
            ["orcid_identifier"],
        ),
        "P2": IsiField(
            "P2",
            "Chapter count (Book Citation Index)",
            _integer,
            ["chapter_count"],
        ),
        "PA": IsiField(
            "PA",
            "Publisher Address",
            functools.partial(_joined, separator="\n"),
            ["publisher_address"],
        ),
        "PD": IsiField("PD", "Publication Date", _joined, ["publication_date"]),
        "PG": IsiField("PG", "Page Count", _integer, ["page_count"]),
        "PI": IsiField("PI", "Publisher City", _joined, ["publisher_city"]),
        "PM": IsiField("PM", "PubMed ID", _joined, ["pubmed_id"]),
        "PN": IsiField("PN", "Part Number", _joined, ["part_number"]),
        "PT": IsiField(
            "PT",
            "Publication Type (J=Journal; B=Book; S=Series; P=Patent)",
            _joined,
            ["publication_type"],
        ),
        "PU": IsiField("PU", "Publisher", _joined, ["publisher"]),
        "PY": IsiField(
            "PY",
            "Year Published",
            _integer,
            ["year_published", "year", "publication_year"],
        ),
        "RI": IsiField(
            "RI", "ResearcherID Number", _delimited, ["researcherid_number"]
        ),
        "RP": IsiField("RP", "Reprint Address", _joined, ["reprint_address"]),
        "SC": IsiField("SC", "Research Areas", _delimited, ["research_areas"]),
        "SE": IsiField("SE", "Book Series Title", _joined, ["book_series_title"]),
        "SI": IsiField("SI", "Special Issue", _joined, ["special_issue"]),
        "SN": IsiField(
            "SN",
            "International Standard Serial Number (ISSN)",
            _joined,
            ["issn"],
        ),
        "SO": IsiField("SO", "Publication Name", _joined, ["publication_name"]),
        "SP": IsiField(
            "SP",
            "Conference Sponsors",
            functools.partial(_delimited, delimiter=", "),
            ["conference_sponsors"],
        ),
        "SU": IsiField("SU", "Supplement", _joined, ["supplement"]),
        "TC": IsiField(
            "TC",
            "Web of Science Core Collection Times Cited Count",
            _integer,
            ["wos_times_cited_count", "wos_times_cited"],
        ),
        "TI": IsiField("TI", "Document Title", _joined, ["title"]),
        "U1": IsiField("U1", "Usage Count (Last 180 Days)", _integer, ["usage_count"]),
        "U2": IsiField("U2", "Usage Count (Since 2013)", _integer, ["usage_count"]),
        "UT": IsiField(
            "UT",
            "Unique Article Identifier",
            _joined,
            ["unique_article_identifier"],
        ),
        "VL": IsiField("VL", "Volume", _joined, ["volume"]),
        "WC": IsiField(
            "WC",
            "Web of Science Categories",
            _delimited,
            ["web_of_science_categories"],
        ),
        "Z9": IsiField(
            "Z9",
            "Total Times Cited Count (WoS Core, BCI, and CSCD)",
            _integer,
            ["total_times_cited_count", "times_cited"],
        ),
    }

    def __init__(self, *isi_files: TextIO):
        self._files = isi_files
        for file in self._files:
            file.seek(0)

    def build(self) -> Collection:
        articles = self._get_articles_from_files()
        return Collection(list(articles))

    def _get_articles_as_str_from_files(self) -> Iterable[str]:
        for file in self._files:
            articles_as_str = file.read().split("\n\n")
            for article_as_str in articles_as_str:
                if article_as_str != "ER" and article_as_str:
                    # Strip `\n` at the end of the article so we don't trip
                    yield article_as_str.strip()

    def _get_articles_from_files(self) -> Iterable[Article]:
        for article_as_str in self._get_articles_as_str_from_files():
            with suppress(MissingCriticalInformation):
                article = self._parse_article_from_str(article_as_str)
                yield article

    @classmethod
    def _get_articles_from_references(
        cls, references: Optional[List[str]]
    ) -> Iterable[Article]:
        if not references:
            return
        for ref_str in references:
            with suppress(InvalidIsiReference):
                yield cls._parse_reference_from_str(ref_str)

    @classmethod
    def _parse_article_from_str(cls, article_as_str: str) -> Article:
        article_data = collections.defaultdict(list)
        article_data.setdefault("CR", [])
        field = None
        for line in article_as_str.split("\n"):
            match = cls.ISI_LINE_PATTERN.match(line)
            if not match:
                raise InvalidIsiLineError(line)
            parsed = match.groupdict()
            field = parsed.get("field") or field
            if not field or "value" not in parsed or parsed["value"] is None:
                continue
            article_data[field].append(parsed["value"])

        processed = cls._parse_all(dict(article_data))

        if not processed.get("year") or not processed.get("authors"):
            raise MissingCriticalInformation()

        return Article(
            authors=processed["authors"],
            year=processed["year"],
            title=processed.get("title"),
            journal=processed.get("source_abbreviation"),
            volume=processed.get("volume"),
            issue=processed.get("issue"),
            page=processed.get("beginning_page"),
            doi=processed.get("DOI"),
            references=list(
                cls._get_articles_from_references(processed.get("references"))
            ),
            keywords=processed.get("keywords"),
            extra=processed,
            sources={article_as_str},
        )

    @classmethod
    def _parse_reference_from_str(cls, reference: str) -> Article:
        match = cls.ISI_CITATION_PATTERN.match(reference)
        if not match:
            raise InvalidIsiReference(reference)
        data = {key: [value] for key, value in match.groupdict().items() if value}
        processed = cls._parse_all(data)
        return Article(
            _label=reference,
            title=processed.get("title"),
            authors=processed.get("authors", []),
            # FIXME: Year is required here
            year=processed.get("year", 1999),
            journal=processed.get("source_abbreviation"),
            volume=processed.get("volume"),
            page=processed.get("beginning_page"),
            doi=processed.get("DOI"),
            extra=processed,
            sources={reference},
        )

    @staticmethod
    def _parse_all(article_data: Dict[str, List[str]]) -> Mapping[str, Any]:
        processed_data = {}
        for key, values in article_data.items():
            parsed_values_dict = IsiCollectionBuilder._parse(key, values)
            processed_data.update(parsed_values_dict)
        return processed_data

    @staticmethod
    def _parse(key: str, value: List[str]) -> Dict:
        if key in {"FN", "VR", "ER"}:
            return {}

        if key in IsiCollectionBuilder.FIELDS:
            field = IsiCollectionBuilder.FIELDS[key]
            parsed_value = field.parse(value)
            return {new_key: parsed_value for new_key in [field.key, *field.aliases]}

        logger.info(f"Found an unknown field with key {key} and value {value}")
        return {key: _ident(value)}



from bibx._entities.collection import Collection
from bibx._entities.collection_builders.base import CollectionBuilder


class GenericCollectionBuilder(CollectionBuilder):
    def __init__(self, *collections: Collection):
        self._collections = collections

    def build(self) -> Collection:
        return Collection([])
